<?php
/**
 * EOrderPositionBehavior class file.
 *
 * @author Ramil Gallyamow <gallyamow@gmail.com>
 * @license http://www.yiiframework.com/license/
 *
 */

/**
 * EOrderPositionBehavior позволяет автоматически генерировать и сохранять
 * значение поля сортировки записей.
 *
 * <pre>
 * public function behaviors()
 * {
 *		return array(
 *			'EOrderPosition'=>array(
 *				'class'=>'ext.EOrderPositionBehavior',
 *				'positionField'=>'order_position', // sorting field name
 *				'conditionFields'=>array('section_id'), // additional conditions fields
 *			)
 *		);
 * }
 * </pre>
 *
 * Имя поля задается в {@link positionField}.
 * Список полей которые должны быть добавлены в условия можно задать
 * в {@link $conditionFields}.
 *
 * В случае использования в больших таблицах, можно отключить вычисление
 * следующего значения поля и использовать вместо режим {@link MODE_PRIMARY}.
 *
 * В случае использования дополнительных условий, особенно в больших таблицах,
 * желательно присутствие индекса для выборки по этим полям.
 */
class EOrderPositionBehavior extends CActiveRecordBehavior
{
	/**
	 * Шаг инкремента поля
	 */
	const INCREMENT_STEP=10;

	/**
	 * Режим при котором значение не будет вычисляться, и вместо него будет
	 * взято значение поля с primary key.
	 */
	const MODE_PRIMARY='primary';

	/**
	 * Режим при котором  значение будет вычисляться но,
	 * по возможности будет применены оптимизации.
	 */
	const MODE_CALCULATION='calculation';

	/**
	 * Режим при котором значение вычисляется всегда.
	 */
	const MODE_FORCE_CALCULATION='force_calculation';

	/**
	 * Имя поля сортировки в таблице которое будет содержать вычисленное значение.
	 * @var string
	 */
	public $positionField='order_position';

	/**
	 * Список полей-условий.
	 * Например здесь можно перечислить поля section_id, lang_id и т.д., если
	 * требуется генерация в пределах определенного множества, определенного
	 * этими полями.
	 * @var array
	 */
	public $conditionFields=null;

	/**
	 * Режим генерации следующего значения поля {@link positionField}.
	 * Может быть равным 'calculation', 'force_calculation' и 'primary'.
	 * В последнем случае значения поля сортировки не вычисляются,
	 * а вместо него берется значение поля c primary key.
	 * @var string
	 */
	public $mode=self::MODE_CALCULATION;

	/**
	 * Определяем следующее значение для поля сортировки.
	 * @param type $event
	 */
	public function afterSave($event)
	{
		parent::afterSave($event);

		if (!in_array($this->mode, array(self::MODE_CALCULATION, self::MODE_PRIMARY, self::MODE_FORCE_CALCULATION)))
			throw new CException('Invalid mode value.');

		if (!$this->owner->hasAttribute($this->positionField))
			throw new CException('The model does not have the property specified in positionField.');

		// устанавливаем новое значение если оно не было задано
		if (empty($this->owner->{$this->positionField}))
		{
			$newPosition=$this->calculateNextValue();
			$this->owner->{$this->positionField}=$newPosition;
			$this->owner->updateByPk($this->owner->getPrimaryKey(), array("{$this->positionField}"=>$newPosition));
		}
	}

	/**
	 * Вычисление следующего значения поля сортировки.
	 */
	private function calculateNextValue()
	{
		$primaryKey=$this->owner->getPrimaryKey();
		$hasOnePrimary=(!is_array($primaryKey) && !is_null($primaryKey));

		// если модель имеет не составной primary ключ
		if ($hasOnePrimary)
		{
			// режим 'primary'
			if ($this->mode==='primary')
				return $primaryKey;

			// если дополнительных условий нет и модель имеет простой primary key,
			// тогда для оптмизации процесса в качестве вычисляемого значения
			// можно взять значение поля с primary key умноженного на
			// шаг инкремента INCREMENT_STEP
			if (is_null($this->conditionFields) && $this->mode!=self::MODE_FORCE_CALCULATION)
				return $primaryKey * self::INCREMENT_STEP;

		}

		// формируем область поиска
		$whereQuery=array(); $whereParams=array();
		if (!is_null($this->conditionFields))
		{
			foreach ($this->conditionFields as $field)
			{
				$whereQuery[]="{$field}=:{$field}";
				$whereParams[":{$field}"]=$this->owner->$field;
			}
		}

		// для быстродействия
		$max=Yii::app()->db->createCommand()
			->select("MAX({$this->positionField}) as max")
			->from($this->owner->tableName())
			->where(implode(' AND ', $whereQuery), $whereParams)
			->queryScalar();

		// округленное до ближайшего целого
		return ceil(($max + self::INCREMENT_STEP) / self::INCREMENT_STEP) * self::INCREMENT_STEP;
	}
}
